"""
Fetch metadata from remote .conda or .tar.bz2 package.

Try to fetch less than the whole file if possible.

This module should only be used to make *partial* reads against a remote
package, typically just the ``info`` portion. If a full ``.conda`` format
package is needed, it is more efficient to download locally first and then use
the file-based API.
"""

import logging
import sys
import urllib.parse
from pathlib import Path

import requests

from . import package_streaming

# Excellent HTTP Range request file-like object
from .lazy_wheel import LazyConda

log = logging.getLogger(__name__)

session = requests.Session()
session.headers["User-Agent"] = "conda-package-streaming/0.1.0"


METADATA_CHECKLIST = frozenset({"info/index.json", "info/recipe/meta.yaml"})


def extract_conda_info(url, destdir, checklist=METADATA_CHECKLIST, session=session):
    """
    Extract info/index.json and info/recipe/meta.yaml from url to destdir; close
    url as soon as those files are found.
    """
    checklist = set(checklist)
    stream = stream_conda_info(url, session=session)
    for tar, member in stream:
        if member.name in checklist:
            tar.extract(member, destdir)
            checklist.remove(member.name)
        if not checklist:
            stream.close()
            # next iteraton of for loop raises GeneratorExit in stream


def stream_conda_info(url, session=session):
    """
    Yield (tar, member) for conda package at url

    Just "info/" for .conda, all members for tar.
    """
    filename, conda = conda_reader_for_url(url, session=session)

    try:
        yield from package_streaming.stream_conda_info(filename, conda)
    finally:
        if hasattr(conda, "release_conn"):
            # For .tar.bz2. Take extra care to drop connections after we are
            # done reading a partial response.
            conda.release_conn()
        conda.close()


def conda_reader_for_url(
    url, session: requests.Session = session, fall_back_to_full_download: bool = False
):
    """
    Return (name, file_like) suitable for package_streaming APIs

    :param url: URL to conda package
    :param session: The session to use for web requests.
    :param fall_back_to_full_download: If true, fall back to downloading the entire
    package instead of streaming it in the case of a specific issue where the server
    incorrectly responds with 416 (Range Not Satisfiable). See LazyZipOverHTTP for
    more details.
    """
    parsed_url = urllib.parse.urlparse(url)
    *_, filename = parsed_url.path.rsplit("/", 1)
    if filename.endswith(".conda"):
        file_id = filename[: -len(".conda")]
        conda = LazyConda(
            url, session, fall_back_to_full_download=fall_back_to_full_download
        )
        conda.prefetch(file_id)
    elif filename.endswith(".tar.bz2"):
        response = session.get(url, stream=True, headers={"Connection": "close"})
        conda = response.raw
    else:
        raise ValueError("Unsupported extension %s", url)
    return filename, conda


if __name__ == "__main__":  # pragma: no cover
    import logging

    logging.basicConfig(level=logging.DEBUG)
    extract_conda_info(sys.argv[1], Path(sys.argv[2]).absolute())



import json
import logging
import os.path
import shutil
import subprocess
from pathlib import Path

import pytest
import server

from conda_package_streaming.transmute import transmute_tar_bz2

log = logging.getLogger(__name__)


LIMIT_TEST_PACKAGES = 16


def find_packages_dirs() -> Path:
    """
    Ask conda for package directories.
    """
    conda_info = json.loads(
        subprocess.run(
            [os.environ["CONDA_EXE"], "info", "--json"],
            stdout=subprocess.PIPE,
            check=True,
        ).stdout
    )

    # XXX can run individual environment's conda (base conda is more likely to
    # have useful cached packages)
    pkgs_dirs = conda_info["pkgs_dirs"] + [os.path.expanduser("~/miniconda3/pkgs")]

    log.debug("search %s", pkgs_dirs)

    first_pkg_dir = next(path for path in pkgs_dirs if os.path.exists(path))

    return Path(first_pkg_dir)


@pytest.fixture(scope="session")
def pkgs_dir(tmp_path_factory):
    """
    Dedicated test package directory.
    """
    return tmp_path_factory.mktemp("pkgs")


@pytest.fixture(scope="session")
def package_server(pkgs_dir, conda_paths):
    thread = server.get_server_thread(pkgs_dir)
    thread.start()
    return thread


@pytest.fixture(scope="session")
def conda_paths(pkgs_dir: Path):
    found_packages = find_packages_dirs()
    conda_paths = []
    for path in found_packages.iterdir():
        if path.name.endswith((".tar.bz2", ".conda")):
            conda_paths.append(path)

    return add_tar_bz2s(conda_paths, pkgs_dir)


def add_tar_bz2s(paths: list[Path], pkgs_dir: Path):
    """
    If there aren't enough .tar.bz2's available, create some from available
    .conda's. Return paths.
    """
    conda_paths: list[Path] = []
    tarbz2_paths: list[Path] = []
    output_paths: list[Path] = []

    assert isinstance(pkgs_dir, Path)

    for path in paths:
        if path.name.endswith(".tar.bz2"):
            tarbz2_paths.append(path)
        elif path.name.endswith(".conda"):
            conda_paths.append(path)

    tarbz2_path: Path = pkgs_dir

    medium_conda_paths = []
    for path in conda_paths:
        if 1 << 20 < path.stat().st_size < 1 << 22:
            medium_conda_paths.append(path)
    medium_conda_paths = medium_conda_paths[:LIMIT_TEST_PACKAGES]

    # this ignores existing .tar.bz2 for simplicity (.tar.bz2 is missing in CI)
    for conda in set(medium_conda_paths + conda_paths[:10]):
        shutil.copy(conda, tarbz2_path)
        transmute_tar_bz2(str(conda), tarbz2_path)

    output_paths.extend(tarbz2_path.glob("*.tar.bz2"))
    output_paths.extend(tarbz2_path.glob("*.conda"))

    return sorted(output_paths)  # sort interleaves .tar.bz2 and .conda



"""
Convert .tar.bz2 to .conda

Uses ``tempfile.SpooledTemporaryFile`` to buffer ``pkg-*.tar`` and
``info-*.tar``, then compress directly into an open `ZipFile` at the end.
`SpooledTemporaryFile` buffers the first 10MB of the package and its metadata in
memory, but writes out to disk for larger packages.
"""

from __future__ import annotations

import os
import tarfile
from collections.abc import Iterator
from pathlib import Path
from typing import Callable

import zstandard

from .create import ZSTD_COMPRESS_LEVEL, ZSTD_COMPRESS_THREADS, conda_builder

# streams everything in .tar.bz2 mode
from .package_streaming import CondaComponent, stream_conda_component


def transmute(
    package,
    path,
    *,
    compressor: Callable[
        [], zstandard.ZstdCompressor
    ] = lambda: zstandard.ZstdCompressor(
        level=ZSTD_COMPRESS_LEVEL, threads=ZSTD_COMPRESS_THREADS
    ),
    is_info: Callable[[str], bool] = lambda filename: filename.startswith("info/"),
) -> Path:
    """
    Convert .tar.bz2 conda package to .conda-format under path.

    :param package: path to .tar.bz2 conda package
    :param path: destination path for transmuted .conda package
    :param compressor: A function that creates instances of
        ``zstandard.ZstdCompressor()`` to override defaults.
    :param is_info: A function that returns True if a file belongs in the
        ``info`` component of a `.conda` package.  ``conda-package-handling``
        (not this package ``conda-package-streaming``) uses a set of regular
        expressions to keep expected items in the info- component, while other
        items starting with ``info/`` wind up in the pkg- component.

    :return: Path to transmuted package.
    """
    assert package.endswith(".tar.bz2"), "can only convert .tar.bz2 to .conda"
    assert os.path.isdir(path)
    stem = os.path.basename(package)[: -len(".tar.bz2")]
    package_stream = stream_conda_component(package)

    return transmute_stream(
        stem,
        path,
        compressor=compressor,
        is_info=is_info,
        package_stream=package_stream,
    )


def transmute_stream(
    stem,
    path,
    *,
    compressor: Callable[
        [], zstandard.ZstdCompressor
    ] = lambda: zstandard.ZstdCompressor(
        level=ZSTD_COMPRESS_LEVEL, threads=ZSTD_COMPRESS_THREADS
    ),
    is_info: Callable[[str], bool] = lambda filename: filename.startswith("info/"),
    package_stream: Iterator[tuple[tarfile.TarFile, tarfile.TarInfo]],
):
    """
    Convert (TarFile, TarInfo) iterator like those produced by
    ``stream_conda_component`` to .conda-format under path. Allows for more
    creative data sources.

    e.g. recompress ``.conda``:

    .. code-block:: python

        transmute_stream(..., package_stream=itertools.chain(
            stream_conda_component("package.conda",
            component=CondaComponent.pkg),
            stream_conda_component("package.conda",
            component=CondaComponent.info),
        ))

    This example could move files between the ``pkg-`` and ``info-`` components
    depending on the ``is_info`` function.

    :param stem: output filename without extension
    :param path: destination path for transmuted .conda package
    :param compressor: A function that creates instances of
        ``zstandard.ZstdCompressor()`` to override defaults.
    :param is_info: A function that returns True if a file belongs in the
        ``info`` component of a `.conda` package.  ``conda-package-handling``
        (not this package ``conda-package-streaming``) uses a set of regular
        expressions to keep expected items in the info- component, while other
        items starting with ``info/`` wind up in the pkg- component.
    :param package_stream: Iterator of (Tarfile, TarInfo) tuples.

    :return: Path to transmuted package.
    """
    output_path = Path(path, f"{stem}.conda")
    with conda_builder(stem, path, compressor=compressor, is_info=is_info) as conda_tar:
        for tar, member in package_stream:
            if member.isfile():
                conda_tar.addfile(member, tar.extractfile(member))
            else:
                conda_tar.addfile(member)

    return output_path


def transmute_tar_bz2(
    package: str,
    path,
) -> Path:
    """
    Convert .conda package to .tar.bz2 format under path.

    Can recompress .tar.bz2 packages.

    Args:
        package: path to .conda or .tar.bz2 package.
        path: destination path for transmuted package.

    Returns:
        Path to transmuted package.
    """
    assert package.endswith((".tar.bz2", ".conda")), "Unknown extension"
    assert os.path.isdir(path)

    incoming_format = ".conda" if package.endswith(".conda") else ".tar.bz2"

    file_id = os.path.basename(package)[: -len(incoming_format)]

    if incoming_format == ".conda":
        # .tar.bz2 MUST place info/ first.
        components = [CondaComponent.info, CondaComponent.pkg]
    else:
        # .tar.bz2 doesn't filter by component
        components = [CondaComponent.pkg]

    output_path = Path(path, f"{file_id}.tar.bz2")

    with open(package, "rb") as fileobj, tarfile.open(output_path, "x:bz2") as pkg_tar:
        for component in components:
            stream = iter(stream_conda_component(package, fileobj, component=component))
            for tar, member in stream:
                if member.isfile():
                    pkg_tar.addfile(member, tar.extractfile(member))
                else:
                    pkg_tar.addfile(member)

    return output_path



import io
import os
import stat
import tarfile
from errno import ELOOP
from pathlib import Path

import pytest

from conda_package_streaming import exceptions, extract, package_streaming

HAS_TAR_FILTER = hasattr(tarfile, "tar_filter")
MAX_CONDAS = 8


def test_extract_stream(conda_paths, tmp_path):
    for i, package in enumerate(conda_paths):
        print(package)
        with open(package, "rb") as fileobj:
            stream = package_streaming.stream_conda_component(
                package, fileobj, component=package_streaming.CondaComponent.pkg
            )
            dest_dir = tmp_path / package.name
            extract.extract_stream(stream, dest_dir)

        if i >= MAX_CONDAS:
            break


def test_extract_all(conda_paths, tmp_path):
    for i, package in enumerate(conda_paths):
        print(package)
        dest_dir = tmp_path / package.name
        extract.extract(package, dest_dir=dest_dir)

        if i >= MAX_CONDAS:
            break


def empty_tarfile(name, mode=0o644, tar_mode="w", create_subdir=False):
    """
    Return BytesIO containing a tarfile with one empty file named :name
    """
    tar = io.BytesIO()
    t = tarfile.open(mode=tar_mode, fileobj=tar)
    if create_subdir:
        tarinfo = tarfile.TarInfo(name=name)
        # Add execute bit for directory
        tarinfo.mode = mode | 0o111
        tarinfo.type = tarfile.DIRTYPE
        t.addfile(tarinfo, io.BytesIO())
        tarinfo = tarfile.TarInfo(name=str(Path(name, name)))
        tarinfo.mode = mode
        t.addfile(tarinfo, io.BytesIO())
    else:
        tarinfo = tarfile.TarInfo(name=name)
        tarinfo.mode = mode
        t.addfile(tarinfo, io.BytesIO())
    t.close()
    tar.seek(0)
    return tar


def not_unicode_tarbz2(
    name=b"\x80\x81".decode("utf-8", errors="surrogateescape"), mode=0o644
):
    """
    Return BytesIO containing a tarfile with one empty file named :name
    """
    return empty_tarfile(name=name, tar_mode="w:bz2")


def test_oserror(tmp_path):
    """
    Fail if tarfile raises OSError (formerly known as IOError)
    """
    tar = empty_tarfile("empty-test")

    class TarELOOP(tarfile.TarFile):
        def extractall(self, path=None, members=None, filter=None):
            raise OSError(ELOOP, "case sensitivity")

    class TarOSError(tarfile.TarFile):
        def extractall(self, path=None, members=None, filter=None):
            raise OSError("not eloop")

    def stream(cls):
        yield (cls(fileobj=tar), tarfile.TarInfo())

    with pytest.raises(exceptions.CaseInsensitiveFileSystemError):
        extract.extract_stream(stream(TarELOOP), tmp_path)

    with pytest.raises(OSError):
        extract.extract_stream(stream(TarOSError), tmp_path)


def stream(fileobj):
    """
    Like the tuples produced by part of conda-package-streaming.
    """
    yield (package_streaming.TarfileNoSameOwner(fileobj=fileobj), tarfile.TarInfo())


def stream_stdlib(fileobj):
    """
    Like the tuples produced by part of conda-package-streaming.
    """
    yield (tarfile.TarFile(fileobj=fileobj), tarfile.TarInfo())


def test_slip(tmp_path):
    """
    Fail if tarfile tries to put files outside its dest_dir (tmp_path)
    """

    tar = empty_tarfile(name="../slip")

    with pytest.raises(exceptions.SafetyError):
        extract.extract_stream(stream(tar), tmp_path)

    # If we are using tarfile.filter, the leading / will be stripped instead.
    tar2 = empty_tarfile(name="/absolute")

    with pytest.raises(exceptions.SafetyError):
        extract.extract_stream(stream(tar2), tmp_path)


def test_chown(conda_paths, tmp_path, mocker):
    for package in conda_paths[:2]:
        print(package)
        with open(package, "rb") as fileobj:
            stream = package_streaming.stream_conda_component(
                package, fileobj, component=package_streaming.CondaComponent.pkg
            )

            for tar, member in stream:
                assert isinstance(tar, package_streaming.TarfileNoSameOwner), tar
                break


@pytest.mark.parametrize(
    "tar_filter",
    (pytest.param(None, id="no tar filter"), pytest.param("data", id="data_filter")),
)
def test_umask(tmp_path, mocker, tar_filter):
    """
    Demonstrate that umask-respecting tar implementation works.

    Mock umask in case it is different on your system.
    """
    if tar_filter is not None and not HAS_TAR_FILTER:
        pytest.skip("Requires tar_filter")
    try:
        MOCK_UMASK = 0o022
        current_umask = os.umask(MOCK_UMASK)
        mocker.patch("conda_package_streaming.package_streaming.UMASK", new=MOCK_UMASK)

        assert (
            package_streaming.TarfileNoSameOwner(
                fileobj=empty_tarfile("file.txt")
            ).umask
            == MOCK_UMASK
        )

        # [
        #   ('S_IFREG', 32768),
        #   ('UF_HIDDEN', 32768),
        #   ('FILE_ATTRIBUTE_INTEGRITY_STREAM', 32768)
        # ]

        # Of the high bits 100755 highest bit 1 can mean just "is regular file"

        name = "naughty_umask"
        tar3 = empty_tarfile(name=name, mode=0o777, create_subdir=True)

        stat_check = stat.S_IRGRP
        stat_name = "S_IRGRP"

        root_path = tmp_path / "stdlib"
        root_path.mkdir()
        files_to_check = [root_path / name, root_path / name / name]

        extract.extract_stream(stream_stdlib(tar3), root_path, tar_filter=tar_filter)
        for file in files_to_check:
            mode = file.stat().st_mode
            # is the new .extractall(filter=) erasing "stat_name"?
            assert mode & stat_check, f"{file} has {stat_name}? %o != %o" % (
                mode,
                mode & stat_check,
            )

        # specifically forbid that stat bit
        MOCK_UMASK |= stat_check
        mocker.patch("conda_package_streaming.package_streaming.UMASK", new=MOCK_UMASK)
        os.umask(MOCK_UMASK)

        root_path = tmp_path / "cps"
        root_path.mkdir()
        files_to_check = [root_path / name, root_path / name / name]

        tar3.seek(0)
        extract.extract_stream(stream(tar3), root_path, tar_filter=tar_filter)
        for file in files_to_check:
            mode = file.stat().st_mode
            if mode & stat_check:
                assert not (mode & stat_check), (
                    f"{file}: No {stat_name} due to umask? %o != %o"
                    % (
                        mode,
                        mode & stat_check,
                    )
                )
    finally:
        os.umask(current_umask)


def test_encoding():
    """
    Some users do not have "utf-8" as the default sys.getfilesystemencoding() or
    sys.getdefaultencoding(). Instead of trying to change the system encoding,
    we prove that stream_conda_component honors the new passed-in encoding which
    is now "utf-8" by default.
    """

    tar = not_unicode_tarbz2()

    # Use new default encoding of "utf-8" regardless of what the system says.
    stream = package_streaming.stream_conda_component(
        "package.tar.bz2", tar, component="pkg"
    )

    with pytest.raises(UnicodeEncodeError):
        for t, member in stream:
            member.name.encode("utf-8")
            print(t, member)

    tar.seek(0)

    # Prove that we are passing encoding all the way down to the TarFile() used
    # for extraction.
    stream = package_streaming.stream_conda_component(
        "package.tar.bz2", tar, component="pkg", encoding="latin-1"
    )

    for t, member in stream:
        member.name.encode("utf-8")
        print(t, member)



"""
Allow conda_package_streaming to work in .tar.bz2-only mode if zstandard is not
available (please immediately install zstandard if this is the case).
"""

import importlib
import sys
import tarfile
import zipfile
from pathlib import Path

import pytest


def test_degraded(tmpdir):
    try:
        sys.modules["zstandard"] = None  # type: ignore

        import conda_package_streaming.extract
        import conda_package_streaming.package_streaming

        importlib.reload(conda_package_streaming.package_streaming)

        testconda = Path(tmpdir, "testconda.conda")
        with zipfile.ZipFile(testconda, "w"):
            pass

        testtar = Path(tmpdir, "test.tar.bz2")
        with tarfile.open(testtar, "w:bz2") as tar:
            tar.addfile(tarfile.TarInfo(name="jim"))

        for (
            tar,
            _,
        ) in conda_package_streaming.package_streaming.stream_conda_component(testtar):
            pass

        with pytest.raises(RuntimeError):
            for (
                tar,
                _,
            ) in conda_package_streaming.package_streaming.stream_conda_component(
                testconda
            ):
                pass  # pragma: no cover

        with pytest.raises(RuntimeError):
            conda_package_streaming.extract.extract(testconda, tmpdir)

    finally:
        sys.modules.pop("zstandard", None)

        import conda_package_streaming.package_streaming

        importlib.reload(conda_package_streaming.package_streaming)
        assert conda_package_streaming.package_streaming.zstandard



import io
import tempfile
from contextlib import closing, contextmanager
from pathlib import Path
from unittest.mock import MagicMock, patch
from zipfile import ZipFile

import pytest
from requests import HTTPError, Session

from conda_package_streaming import lazy_wheel
from conda_package_streaming.lazy_wheel import LazyConda
from conda_package_streaming.url import (
    conda_reader_for_url,
    extract_conda_info,
    stream_conda_info,
)

LIMIT = 16


@pytest.fixture
def package_url(package_server):
    """
    Base url for all test packages.
    """
    host, port = package_server.server.server_address
    return f"http://{host}:{port}/pkgs"


@pytest.fixture
def package_urls(package_server, package_url):
    pkgs_dir = Path(package_server.app.pkgs_dir)
    conda = []
    tar_bz2 = []

    for path in pkgs_dir.iterdir():
        if len(conda) > LIMIT and len(tar_bz2) > LIMIT:
            break
        url = f"{package_url}/{path.name}"
        if path.name.endswith(".tar.bz2") and len(tar_bz2) < LIMIT:
            tar_bz2.append(url)
        elif path.name.endswith(".conda") and len(conda) < LIMIT:
            conda.append(url)
    # interleave
    urls = []
    for pair in zip(conda, tar_bz2):
        urls.extend(pair)
    return urls


def test_stream_url(package_urls):
    with pytest.raises(ValueError):
        next(stream_conda_info("https://localhost/notaconda.rar"))

    for url in package_urls:
        with closing(stream_conda_info(url)) as members:
            print("stream_url", url)
            for tar, member in members:
                if member.name == "info/index.json":
                    break
            else:
                pytest.fail("info/index.json not found")


def test_fetch_meta(package_urls):
    for url in package_urls:
        with tempfile.TemporaryDirectory() as destdir:
            extract_conda_info(url, destdir)


def test_lazy_wheel(package_urls):
    lazy_tests = 7
    for url in package_urls:
        if url.endswith(".conda"):
            # API works with `.tar.bz2` but only returns LazyConda for `.conda`
            filename, conda = conda_reader_for_url(url)
            assert filename == url.rsplit("/")[-1]
            with conda:
                assert isinstance(conda, LazyConda)
                assert conda.mode == "rb"
                assert conda.readable()
                assert not conda.writable()
                assert not conda.closed

                request_count = conda._request_count

                # did we really prefetch the info?
                zf = ZipFile(conda)  # type: ignore
                filename = filename[: -len(".conda")]
                zf.open(f"info-{filename}.tar.zst").read()

                assert conda._request_count == request_count, (
                    "info required extra GET request"
                )
                assert conda._request_count <= 3

                conda.prefetch("not-appearing-in-archive.txt")

                # zip will figure this out naturally; delete method?
                conda._check_zip()

            lazy_tests -= 1
            if lazy_tests <= 0:
                break
    else:
        raise LookupError(
            f"not enough .conda packages found {lazy_tests} {package_urls}"
        )

    with pytest.raises(HTTPError):
        conda_reader_for_url(package_urls[0] + ".404.conda")

    class Session200(Session):
        def get(self, *args, **kwargs):
            response = super().get(*args, **kwargs)
            response.status_code = 200
            return response

    with pytest.raises(lazy_wheel.HTTPRangeRequestUnsupported):
        LazyConda(package_urls[0], Session200())

    for url in package_urls:
        if url.endswith(".tar.bz2"):
            LazyConda(url, Session())._check_zip()
            break
    else:
        raise LookupError("no .tar.bz2 packages found")


@pytest.mark.parametrize("fall_back_to_full_download", [True, False])
@patch("conda_package_streaming.url.LazyConda")
def test_conda_reader_for_url_passes_to_lazy_conda_correctly(
    lazy_conda_mock: MagicMock, fall_back_to_full_download: bool
):
    url = "https://example.com/package.conda"
    session = Session()

    filename, conda = conda_reader_for_url(
        url, session, fall_back_to_full_download=fall_back_to_full_download
    )

    assert filename == "package.conda"
    lazy_conda_mock.assert_called_once_with(
        url,
        session,
        fall_back_to_full_download=fall_back_to_full_download,
    )


def test_no_file_after_info():
    """
    If info is the last file, LazyConda must fetch (start of info file .. start
    of zip directory) instead of to the next file in the zip.
    """

    class MockBytesIO(io.BytesIO):
        prefetch = LazyConda.prefetch

        @contextmanager
        def _stay(self):
            yield

    zip = MockBytesIO()
    zf = ZipFile(zip, "w")
    zf.writestr("info-test.tar.zst", b"00000000")  # a short file
    zf.close()

    zip.prefetch("test")


@pytest.mark.skip()
def test_obsolete_lazy_wheel_selftest():
    import logging

    import requests

    logging.basicConfig(level=logging.DEBUG)

    session = requests.Session()

    lzoh = lazy_wheel.LazyZipOverHTTP(
        "https://repodata.fly.dev/repo.anaconda.com/pkgs/main/win-32/current_repodata.jlap",
        session,
    )

    lzoh.seek(1024)
    lzoh.read(768)
    lzoh.seek(0)

    # compare against regular fetch
    with open("outfile.txt", "wb+") as out:
        buf = b" "
        while buf:
            buf = lzoh.read(1024 * 10)
            print(list(zip(lzoh._left, lzoh._right)), lzoh._length)
            if not buf:
                break
            out.write(buf)



"""
Tools for creating ``.conda``-format archives.

Uses ``tempfile.SpooledTemporaryFile`` to buffer ``pkg-*.tar`` and
``info-*.tar``, then compress directly into an open `ZipFile` at the end.
`SpooledTemporaryFile` buffers the first 10MB of the package and its metadata in
memory, but writes out to disk for larger packages.

Uses more disk space than ``conda-package-handling`` (temporary uncompressed
tarballs of the package contents) but accepts streams instead of just
files-on-the-filesystem.
"""

from __future__ import annotations

import json
import shutil
import tarfile
import tempfile
import zipfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Callable

import zstandard

# increase to reduce speed and increase compression (levels above 19 use much
# more memory)
ZSTD_COMPRESS_LEVEL = 19
# increase to reduce compression and increase speed
ZSTD_COMPRESS_THREADS = 1

CONDA_PACKAGE_FORMAT_VERSION = 2

# Account for growth from "2 GB of /dev/urandom" to not exceed ZIP64_LIMIT after
# compression
CONDA_ZIP64_LIMIT = zipfile.ZIP64_LIMIT - (1 << 18) - 1


def anonymize(tarinfo: tarfile.TarInfo):
    """
    Pass to ``tarfile.add(..., filter=anonymize)`` to anonymize uid/gid.

    Does not anonymize mtime or any other field.
    """
    tarinfo.uid = tarinfo.gid = 0
    tarinfo.uname = tarinfo.gname = ""
    return tarinfo


class CondaTarFile(tarfile.TarFile):
    """
    Subclass of :external+python:py:class:`tarfile.TarFile` that adds members to
    a second ``info`` tar if they match ``is_info(name)``.

    Create this with ``conda_builder(...)`` which sets up the component
    archives, then wraps them into a ``.conda`` on exit.

    Only useful for creating, not extracting ``.conda``.
    """

    info_tar: tarfile.TarFile
    is_info: Callable

    def __init__(
        self,
        *args,
        info_tar: tarfile.TarFile,
        is_info=lambda name: name.startswith("info/"),
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.info_tar = info_tar
        self.is_info = is_info

    def addfile(self, tarinfo, fileobj=None):
        """
        Add the TarInfo object ``tarinfo`` to the archive. If ``fileobj`` is
        given, it should be a binary file, and tarinfo.size bytes are read from
        it and added to the archive. You can create TarInfo objects directly, or
        by using ``gettarinfo()``.

        If ``self.is_info(tarinfo.name)`` returns ``True``, add ``tarinfo`` to
        ``self.info_tar`` instead.
        """
        if self.is_info(tarinfo.name):
            return self.info_tar.addfile(tarinfo, fileobj=fileobj)
        else:
            return super().addfile(tarinfo, fileobj)


@contextmanager
def conda_builder(
    stem,
    path,
    *,
    compressor: Callable[
        [], zstandard.ZstdCompressor
    ] = lambda: zstandard.ZstdCompressor(
        level=ZSTD_COMPRESS_LEVEL, threads=ZSTD_COMPRESS_THREADS
    ),
    is_info: Callable[[str], bool] = lambda filename: filename.startswith("info/"),
    encoding="utf-8",
) -> Iterator[CondaTarFile]:
    """
    Produce a ``TarFile`` subclass used to build a ``.conda`` package. The
    subclass delegates ``addfile()`` to the ``info-`` component when ``is_info``
    returns True.

    When the context manager exits, ``{path}/{stem}.conda`` is written with
    the component tar archives.

    Args:
        stem: output filename without extension

        path: destination path for transmuted .conda package compressor: A
            function that creates instances of ``zstandard.ZstdCompressor()``.

        encoding: passed to TarFile constructor. Keep default "utf-8" for valid
            .conda.

    Yields:
        ``CondaTarFile``
    """
    output_path = Path(path, f"{stem}.conda")
    with (
        tempfile.SpooledTemporaryFile() as info_file,
        tempfile.SpooledTemporaryFile() as pkg_file,
    ):
        with (
            tarfile.TarFile(fileobj=info_file, mode="w", encoding=encoding) as info_tar,
            CondaTarFile(
                fileobj=pkg_file,
                mode="w",
                info_tar=info_tar,
                is_info=is_info,
                encoding=encoding,
            ) as pkg_tar,
        ):
            # If we wanted to compress these at a low setting to save temporary
            # space, we could insert a file object that counts bytes written in
            # front of a zstd (level between 1..3) compressor.
            yield pkg_tar

            info_tar.close()
            pkg_tar.close()

            info_size = info_file.tell()
            pkg_size = pkg_file.tell()

            info_file.seek(0)
            pkg_file.seek(0)

        with zipfile.ZipFile(
            output_path,
            "x",  # x to not append to existing
            compresslevel=zipfile.ZIP_STORED,
        ) as conda_file:
            # Use a maximum of one Zstd compressor, stream_writer at a time to save
            # # memory.
            data_compress = compressor()

            pkg_metadata = {"conda_pkg_format_version": CONDA_PACKAGE_FORMAT_VERSION}
            conda_file.writestr("metadata.json", json.dumps(pkg_metadata))

            with (
                conda_file.open(
                    f"pkg-{stem}.tar.zst",
                    "w",
                    force_zip64=(pkg_size > CONDA_ZIP64_LIMIT),
                ) as pkg_file_zip,
                data_compress.stream_writer(
                    pkg_file_zip, size=pkg_size, closefd=False
                ) as pkg_stream,
            ):
                shutil.copyfileobj(pkg_file._file, pkg_stream)

            with (
                conda_file.open(
                    f"info-{stem}.tar.zst",
                    "w",
                    force_zip64=(info_size > CONDA_ZIP64_LIMIT),
                ) as info_file_zip,
                data_compress.stream_writer(
                    info_file_zip,
                    size=info_size,
                    closefd=False,
                ) as info_stream,
            ):
                shutil.copyfileobj(info_file._file, info_stream)



"""Lazy ZIP over HTTP"""

from __future__ import annotations

import logging
import zipfile
from bisect import bisect_left, bisect_right
from collections.abc import Iterator
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from typing import Any
from zipfile import BadZipfile, ZipFile

from requests import HTTPError, Session
from requests.models import CONTENT_CHUNK_SIZE, Response

# from pip 22.0.3 with fixes & remove imports from pip


log = logging.getLogger(__name__)

# If-Match (etag) to detect file changed during fetch would also be nice
HEADERS = {"Accept-Encoding": "identity"}


class HTTPRangeRequestUnsupported(Exception):
    pass


class LazyZipOverHTTP:
    """File-like object mapped to a ZIP file over HTTP.

    This uses HTTP range requests to lazily fetch the file's content,
    which is supposed to be fed to ZipFile.  If such requests are not
    supported by the server, raise HTTPRangeRequestUnsupported
    during initialization.
    """

    def __init__(
        self,
        url: str,
        session: Session,
        chunk_size: int = CONTENT_CHUNK_SIZE,
        fall_back_to_full_download: bool = False,
    ) -> None:
        """
        Initialize a LazyZipOverHTTP object.

        :param session: The session to use for web requests.
        :param chunk_size: The chunk size to use for downloading.
        :param fall_back_to_full_download: If true, we fall back to downloading
        the whole file if the server incorrectly responds with 416
        (Range Not Satisfiable) to an HTTP range request.
        """
        # if CONTENT_CHUNK_SIZE is bigger than the file:
        # In [8]: response.headers["Content-Range"]
        # Out[8]: 'bytes 0-3133374/3133375'

        self._request_count = 0

        self._session, self._url, self._chunk_size = session, url, chunk_size

        self._fall_back_to_full_download: bool = fall_back_to_full_download
        self._has_streaming_support: bool = True

        # If the server returns 416 (Range Not Satisfiable) and the fallback is
        # enabled, we request the whole file and set this to False. Some package
        # servers incorrectly respond with 416 (Range Not Satisfiable) if the
        # file is smaller than the range requested.

        # initial range request for the end of the file
        # if the server does not support range requests, this sets
        # _has_streaming_support to False
        tail = self._stream_response(start="", end=CONTENT_CHUNK_SIZE)
        # e.g. {'accept-ranges': 'bytes', 'content-length': '10240',
        # 'content-range': 'bytes 12824-23063/23064', 'last-modified': 'Sat, 16
        # Apr 2022 13:03:02 GMT', 'date': 'Thu, 21 Apr 2022 11:34:04 GMT'}

        if self._has_streaming_support and tail.status_code != 206:
            raise HTTPRangeRequestUnsupported("range request is not supported")

        if self._has_streaming_support:
            # lowercase content-range to support s3
            self._length = int(tail.headers["content-range"].partition("/")[-1])
        else:
            # the file is already downloaded
            self._length = len(tail.content)

        self._file = NamedTemporaryFile()
        self.truncate(self._length)

        # length is also in Content-Length and Content-Range header
        with self._stay():
            if self._has_streaming_support:
                content_length = int(tail.headers["content-length"])
                if hasattr(tail, "content"):
                    assert content_length == len(tail.content)
            else:
                content_length = len(tail.content)
            self.seek(self._length - content_length)
            for chunk in tail.iter_content(self._chunk_size):
                self._file.write(chunk)
        self._left: list[int] = [self._length - content_length]
        self._right: list[int] = [self._length - 1]

    @property
    def mode(self) -> str:
        """Opening mode, which is always rb."""
        return "rb"

    @property
    def name(self) -> str:
        """Path to the underlying file."""
        return self._file.name

    def seekable(self) -> bool:
        """Return whether random access is supported, which is True."""
        return True

    def close(self) -> None:
        """Close the file."""
        self._file.close()

    @property
    def closed(self) -> bool:
        """Whether the file is closed."""
        return self._file.closed

    def read(self, size: int = -1) -> bytes:
        """Read up to size bytes from the object and return them.

        As a convenience, if size is unspecified or -1,
        all bytes until EOF are returned.  Fewer than
        size bytes may be returned if EOF is reached.
        """
        # BUG does not download correctly if size is unspecified
        download_size = size
        start, length = self.tell(), self._length
        stop = length if size < 0 else min(start + download_size, length)
        start = max(0, stop - download_size)
        self._download(start, stop - 1)
        return self._file.read(size)

    def readable(self) -> bool:
        """Return whether the file is readable, which is True."""
        return True

    def seek(self, offset: int, whence: int = 0) -> int:
        """Change stream position and return the new absolute position.

        Seek to offset relative position indicated by whence:
        * 0: Start of stream (the default).  pos should be >= 0;
        * 1: Current position - pos may be negative;
        * 2: End of stream - pos usually negative.
        """
        return self._file.seek(offset, whence)

    def tell(self) -> int:
        """Return the current position."""
        return self._file.tell()

    def truncate(self, size: int | None = None) -> int:
        """Resize the stream to the given size in bytes.

        If size is unspecified resize to the current position.
        The current stream position isn't changed.

        Return the new file size.
        """
        return self._file.truncate(size)

    def writable(self) -> bool:
        """Return False."""
        return False

    def __enter__(self) -> LazyZipOverHTTP:
        self._file.__enter__()
        return self

    def __exit__(self, *exc: Any) -> bool | None:
        return self._file.__exit__(*exc)

    @contextmanager
    def _stay(self) -> Iterator[None]:
        """Return a context manager keeping the position.

        At the end of the block, seek back to original position.
        """
        pos = self.tell()
        try:
            yield
        finally:
            self.seek(pos)

    def _check_zip(self) -> None:
        """Check and download until the file is a valid ZIP."""
        end = self._length - 1
        for start in reversed(range(0, end, self._chunk_size)):
            self._download(start, end)
            with self._stay():
                try:
                    # For read-only ZIP files, ZipFile only needs
                    # methods read, seek, seekable and tell.
                    ZipFile(self)  # type: ignore
                except BadZipfile:
                    pass
                else:
                    break

    def _stream_response(
        self, start: int | str, end: int, base_headers: dict[str, str] = HEADERS
    ) -> Response:
        """
        Return HTTP response to a range request from start to end.
        If the does not support range requests, the whole file is requested.

        :param start: if "", request `end` bytes from end of file.
        """
        headers = base_headers.copy()
        headers["Range"] = f"bytes={start}-{end}"
        log.debug("%s", headers["Range"])
        # TODO: Get range requests to be correctly cached
        headers["Cache-Control"] = "no-cache"
        self._request_count += 1

        if self._has_streaming_support:
            response = self._session.get(self._url, headers=headers, stream=True)

            if response.status_code == 416 and self._fall_back_to_full_download:
                # Range Not Satisfiable -> enable fallback
                self._has_streaming_support = False
            elif response.status_code == 416:
                # fallback disabled, print helpful error message
                raise HTTPError(
                    "The server returned 416 (Range Not Satisfiable). "
                    "This can occur on a buggy server if the file is smaller "
                    "than the range requested. Set the fall_back_to_full_download flag "
                    "to work around this issue.",
                    response=response,
                )
            else:
                response.raise_for_status()
                return response

        # no streaming support, try to get the whole file
        del headers["Range"]
        response = self._session.get(self._url, headers=headers, stream=True)

        response.raise_for_status()
        return response

    def _merge(
        self, start: int, end: int, left: int, right: int
    ) -> Iterator[tuple[int, int]]:
        """Return an iterator of intervals to be fetched.

        Args:
            start (int): Start of needed interval
            end (int): End of needed interval
            left (int): Index of first overlapping downloaded data
            right (int): Index after last overlapping downloaded data
        """
        lslice, rslice = self._left[left:right], self._right[left:right]
        i = start = min([start] + lslice[:1])
        end = max([end] + rslice[-1:])
        for j, k in zip(lslice, rslice):
            if j > i:
                yield i, j - 1
            i = k + 1
        if i <= end:
            yield i, end
        self._left[left:right], self._right[left:right] = [start], [end]

    def _download(self, start: int, end: int) -> None:
        """
        Download bytes from start to end inclusively. If the server does not support
        streaming for this file, this does nothing as the entire file is already
        downloaded.
        """
        # these assertions should hold, but read() violates them
        # assert 0 <= start <= end
        # assert end < self._length

        with self._stay():
            left = bisect_left(self._right, start)
            right = bisect_right(self._left, end)
            for start, end in self._merge(start, end, left, right):
                response = self._stream_response(start, end)
                self.seek(start)
                for chunk in response.iter_content(self._chunk_size):
                    self._file.write(chunk)


class LazyConda(LazyZipOverHTTP):
    def prefetch(self, conda_file_id):
        """
        Conda fork specific. Prefetch the `.info` range from the remote archive.
        Reduces number of Range requests to 2 or 3 (1 or 2 for the directory, 1
        for the file).

        conda_file_id: name of .conda without path or `.conda` extension
        """
        target_file = f"info-{conda_file_id}.tar.zst"
        with self._stay():  # not strictly necessary
            # try to read entire conda info in one request
            zf = zipfile.ZipFile(self)
            infolist = zf.infolist()
            for i, info in enumerate(infolist):
                if info.filename == target_file:
                    # could be incorrect if zipfile was concatenated to another
                    # file (not likely for .conda)
                    start = info.header_offset
                    try:
                        end = infolist[i + 1].header_offset
                        # or info.header_offset
                        # + len(info.filename)
                        # + len(info.extra)
                        # + info.compress_size
                        # (unless Zip64)
                    except IndexError:
                        end = zf.start_dir
                    self.seek(start)
                    self.read(end - start)
                    log.debug(
                        "prefetch %s-%s",
                        info.header_offset,
                        end,
                    )
                    break
            else:
                log.debug("no zip prefetch")



import re

import pytest
import requests
import responses
from requests import HTTPError
from requests.models import PreparedRequest
from responses import matchers

from conda_package_streaming.lazy_wheel import LazyZipOverHTTP

HTTP_FULL_RANGE_PATTERN = re.compile(r"bytes=(\d+)-(\d+)$")
HTTP_END_RANGE_PATTERN = re.compile(r"bytes=-(\d+)$")


class TestLazyZipOverHTTP:
    @staticmethod
    def generate_zero_bytes(length_bytes: int) -> bytes:
        return bytes(length_bytes)

    @staticmethod
    def successful_http_stream_callback_wrapper(file: bytes):
        # https://datatracker.ietf.org/doc/html/rfc7233
        def _callback(request: PreparedRequest):
            full_pattern_match = HTTP_FULL_RANGE_PATTERN.match(request.headers["Range"])
            end_pattern_match = HTTP_END_RANGE_PATTERN.match(request.headers["Range"])

            if full_pattern_match:
                start_range = int(full_pattern_match.group(1))
                end_range = int(full_pattern_match.group(2))

                assert start_range >= 0
                assert end_range >= 0
                assert start_range < end_range

                if start_range >= len(file):
                    # Range Not Satisfiable
                    return 416, {}, b""

                # truncate to file length
                end_range = min(end_range, len(file) - 1)
                content_length = end_range - start_range + 1  # the bounds are inclusive
            else:
                assert end_pattern_match

                content_length_requested = int(end_pattern_match.group(1))
                assert content_length_requested >= 0
                content_length = min(content_length_requested, len(file))

                start_range = len(file) - content_length
                end_range = len(file) - 1

            headers = {
                "Content-Length": str(content_length),
                "Content-Range": f"bytes {start_range}-{end_range}/{len(file)}",
            }

            # this is not inlined because black and flake8 disagree on how to format it
            end_file_range = end_range + 1

            return 206, headers, file[start_range:end_file_range]

        return _callback

    @pytest.mark.parametrize("fall_back_to_full_download", [True, False])
    @responses.activate
    def test_init_stream_successful(self, fall_back_to_full_download: bool):
        responses.add_callback(
            responses.GET,
            "https://example.com/test.zip",
            callback=self.successful_http_stream_callback_wrapper(
                self.generate_zero_bytes(10000)
            ),
            content_type="application/zip",
        )

        session = requests.Session()
        lazy_zip = LazyZipOverHTTP(
            "https://example.com/test.zip",
            session,
            fall_back_to_full_download=fall_back_to_full_download,
        )
        lazy_zip.read()

    @pytest.mark.parametrize("fall_back_to_full_download", [True, False])
    @responses.activate
    def test_init_stream_retry_without_range_with_fallback(
        self, fall_back_to_full_download: bool
    ):
        """
        Some package servers respond with 416 (Range Not Satisfiable) when the
        file is smaller than the range requested. This violates RFC 7233, but we
        cope with it by retrying without Range, requesting the full file if
        fall_back_to_full_download is set.
        """
        responses.add(
            responses.GET,
            url="https://example.com/test.zip",
            status=416,
            match=[matchers.header_matcher({"Range": re.compile(r".*")})],
        )

        responses.add(
            responses.GET,
            url="https://example.com/test.zip",
            body=self.generate_zero_bytes(10000),
            content_type="application/zip",
        )

        session = requests.Session()

        if fall_back_to_full_download:
            # this should work
            lazy_zip = LazyZipOverHTTP(
                "https://example.com/test.zip", session, fall_back_to_full_download=True
            )
            lazy_zip.read()
            return

        # otherwise, the constructor should raise an exception
        with pytest.raises(
            HTTPError,
            match="Set the fall_back_to_full_download flag to work around this issue.",
        ):
            LazyZipOverHTTP(
                "https://example.com/test.zip",
                session,
                fall_back_to_full_download=False,
            )
